/*
	Project:	UV pcb exposure timer
	Author:		Mario "" Mueh
	Date:		05.12.2011
	Last chgd:	03.09.2012

	Desc:		9:59 exposure timer for two-zone uv exposure device.
	Hardware:	Attiny2313 @ 4MHz
				3-digit LED display, common cathode, multiplexed
				2x SHARP S202* semiconductor relais
				ALPS rotary encoder		

	This program uses some code originally written by Peter Dannegger (published on mikrocontroller.net) to
		read the rotary encoder. For more info, please see http://www.mikrocontroller.net/topic/112603 and
		http://www.mikrocontroller.net/articles/Drehgeber#Solide_L.C3.B6sung:_Beispielcode_in_C
	
*/

/* CPU and speed set as attiny2313 at 4.000.000 Hz in avrstudio 4 config - define, if compiled with other envs! */
/* Compiled with WinAVR avr-gcc using ATMEL AVRStudio 4 */

#include <stdio.h>
#include <string.h>	
#include <stdlib.h>
#include <avr/io.h>	
#include <util/delay.h>
#include <avr/interrupt.h>

/* Define display hardware */

/* Column driver transistors */
#define LD_1 (1<<PD6);
#define LD_2 (1<<PD5);
#define LD_3 (1<<PD4);

/* Define LED characters for segment outputs */

#define LED_MASK 0b01111111;

static const unsigned char segment_map[22] = 
{
	/* Segment order of connection: 0  RU T  LU M  LL RL B  */

	0b01110111,	// 0
	0b01000010, // 1
	0b01101101, // 2
	0b01101011, // 3
	0b01011010, // 4
	0b00111011, // 5
	0b00111111, // 6
	0b01100010, // 7
	0b01111111, // 8
	0b01111011, // 9

	0b01111110, // A
	0b00011111, // B
	0b00110101, // C
	0b01001111, // D
	0b00111101, // E
	0b00111100, // F

	0b01111100, // P
	0b00001100, // R
	0b01011011, // Y
	0b00001110, // N
	0b00000000, // SPACE

	0b00000001 // INVALID CHAR
};


#define DEBOUNCE_TIME 200

/* Global variables */
uint8_t mode;

int8_t uvMinutes;
int8_t uvSeconds;
int8_t uvZone;
int8_t uvPreheat;

uint8_t ledBuffer[3];

volatile uint8_t value;
volatile uint16_t timeDivider;
volatile uint8_t timerTick;
volatile uint16_t timeDivider2;
volatile uint8_t timerTick2;

volatile uint8_t keyDebounce;
volatile uint8_t keyPressed;

#define PHASE_A     (PIND & 1<<PD2)
#define PHASE_B     (PIND & 1<<PD3)

volatile int8_t enc_delta;          // -128 ... 127
static int8_t last;


/* Function prototypes */

void 		waitms(uint16_t ms);
void 		draw(uint8_t t_on, uint8_t t_off);
void 		display(char data[]);

uint8_t 	key_read( void );
int8_t 		encode_read4( void );
void 		encode_init( void );


/* Main program loop */

int main() {

	char buf[6];
	char* bufptr;
	
	// Port setup
	// D0+D1:    Output channels to opto-relais
	// D4+D5+D6: LED block select transistors
	// D2+D3:    Input from rotary encoder
	// B0-6:     LED segment drivers
	// B7:       Input from rotary encoder button
	
	/* Select port directions */
	DDRB = ((1<<PB6) | (1<<PB5) | (1<<PB4) | (1<<PB3) | (1<<PB2) | (1<<PB1) | (1<<PB0));
	DDRD = ((1<<PD6) | (1<<PD5) | (1<<PD4) | (1<<PD1) | (1<<PD0));

	/* Pullup resistors */
	PORTD = ((1<<PD2) | (1<<PD3));
	PORTB = (1<<PB7);

	/* Timer setup for rotary encoder and timekeeping purposes */
	TCCR0B |= (1<<CS01);
	TIMSK |= (1<<TOIE0);

	/* Enable PCINT 7 */
	GIMSK |= (1<<PCIE);
	PCMSK |= (1<<PCINT7);

	/* Init Peter Danneggers rotary encoder module */
	encode_init();

	/* Set to first phase: Waiting for input */
	// mode = 0;
	
	/* Default setup */
	uvMinutes = 2;
	uvSeconds = 45;
	uvZone = 3;
	uvPreheat = 15;

	/* Reset statekeeping vars */
	timeDivider = 0;
	timeDivider2 = 0;
	timerTick2 = 0;
	timerTick = 0;
	mode = 1;
	
	/* Leave some settling time and enable interrupts */
	waitms(100);
	sei();

	// Loop main program forever
	while (1==1) {
		// value += encode_read4();
		PORTD &= ~((1<<PD1) | (1<<PD0));

		/* 1st: Time setting minutes */
		if (mode == 1) {
			/* TODO: Blink first char */
			uvMinutes += encode_read4();
			if (uvMinutes >= 10)
				uvMinutes = 9;
			else if (uvMinutes < 0)
				uvMinutes = 0;

			if (key_read())
				mode = 2;

			buf[0] = '0';
			utoa(uvMinutes, buf, 10);
			buf[1] = '_'; buf[2] = '_';
		}

		/* 2nd: Time setting seconds */
		else if (mode == 2) {
			/* TODO: Blink second+third char */
			PORTD &= ~((1<<PD1) | (1<<PD0));

			uvSeconds += encode_read4();
			if (uvSeconds >= 60)
				uvSeconds = 59;
			else if (uvSeconds < 0)
				uvSeconds = 0;

			if (key_read())
				mode = 3;

			buf[0] = '_';
			buf[1] = '0'; buf[2] = '0';
			bufptr = 1+(char*)buf;
			if (uvSeconds < 10) 
				bufptr++;
			utoa(uvSeconds, bufptr, 10);
		}

		/* 3rd: Preheat time setting seconds */
		else if (mode == 3) {
			/* TODO: Blink second+third char */
			PORTD &= ~((1<<PD1) | (1<<PD0));

			uvPreheat += encode_read4();
			if (uvPreheat >= 60)
				uvPreheat = 59;
			else if (uvPreheat < 0)
				uvPreheat = 0;

			if (key_read())
				mode = 4;

			buf[0] = 'P'; buf[1] = '0'; buf[2] = '0';
			bufptr = 1+(char*)buf;
			if (uvPreheat < 10) 
				bufptr++;
			utoa(uvPreheat, bufptr, 10);
		}

		/* 4th: Active UV areas (none, A, B, both) */
		else if (mode == 4) {
			PORTD &= ~((1<<PD1) | (1<<PD0));
			if (key_read())
				mode = 5;

			buf [0] = 'A'; buf [1] = '1'; buf[2] = '2';
		}

		/* 5th: Ready, press button to start */
		else if (mode == 5) {
			PORTD &= ~((1<<PD1) | (1<<PD0));
			if (key_read())
				mode = 6;

			buf [0] = 'R'; buf [1] = 'D'; buf[2] = 'Y';
		}

		/* 6th: Preheat countdown */
		else if (mode == 6) {
			PORTD |= ((1<<PD1) | (1<<PD0));
			if (timerTick != 0) {
				uvPreheat--;
				if (uvPreheat <= 0)
					mode = 7;
				timerTick = 0;
			}

			buf[0] = 'P'; buf[1] = '0'; buf[2] = '0';
			bufptr = 1+(char*)buf;
			if (uvPreheat < 10) 
				bufptr++;
			utoa(uvPreheat, bufptr, 10);
			// bergang von Preheat auf N_ormal: 1sec fehlt
		}

		/* 7th: Exposure countdown */
		else if (mode == 7) {
			PORTD |= ((1<<PD1) | (1<<PD0));
			if (timerTick != 0) {
				uvSeconds--;
				if (uvSeconds < 0) {
					uvSeconds = 59;
					uvMinutes--;
				}
				if (uvMinutes <= 0 && uvSeconds <= 0)
					mode = 8;
				timerTick = 0;
			}

			buf[0] = '0';
			utoa(uvMinutes, buf, 10);
			buf[1] = '0'; buf[2] = '0';
			bufptr = 1+(char*)buf;
			if (uvSeconds < 10) 
				bufptr++;
			utoa(uvSeconds, bufptr, 10);
		}

		/* 8th: Exposure completed */
		else if (mode == 8) {
			PORTD &= ~((1<<PD1) | (1<<PD0));
			if (key_read())
				mode = 1;

			buf [0] = 'R'; buf [1] = 'D'; buf[2] = 'Y';
		}


		display(buf);
		
		/* Switch between blinking display modes */
		if (mode != 8)
			draw(1,1);
		else if (mode == 8 && timerTick2 != 0) {
			timerTick2 = 0;
			draw(1,1);
		}		
	}

	cli();

	return 0;
}

/* Timer overflow interrupt, entered every 0.5ms */
ISR( TIMER0_OVF_vect ) {
	int8_t new, diff;
 
	new = 0;
	if( PHASE_A )
		new = 3;
	if( PHASE_B )
		new ^= 1;
	diff = last - new;
	if( diff & 1 ){
		last = new;
		enc_delta += (diff & 2) - 1;
	}

	timeDivider++;
	if (mode >= 5 && timeDivider >= 1953) {
		timerTick = 1;
		timeDivider = 0;
	}

	timeDivider2++;
	if (timeDivider2 >= 390) {
		timerTick2 = 1;
		timeDivider2 = 0;
	}

	if (keyDebounce > 0 && !(PINB & (1<<PB7)))
		keyDebounce--;
}

/* Pinchange interrupt */
ISR( PCINT_vect ) {
	if (keyDebounce == 0 && (PINB & (1<<PB7)))
		keyPressed = 1;
		
}

uint8_t key_read( void ) {
	if (keyPressed != 0) {
		keyDebounce = DEBOUNCE_TIME;
		keyPressed = 0;
		return 255;
	}
	return 0;
}

/* Decode graycode */
int8_t encode_read4( void )
{
  int8_t val;
 
  cli();
  val = enc_delta;
  enc_delta = val & 3;
  sei();
  return val >> 2;
}

/* Init rotary encoder detection */
void encode_init( void )
{
  int8_t new;
 
  new = 0;
  if( PHASE_A )
    new = 3;
  if( PHASE_B )
    new ^= 1;
  last = new;
  enc_delta = 0;
}

/* Write preprocessed display data to hardware */
void draw(uint8_t t_on, uint8_t t_off) {
	waitms(t_off);
	PORTD |= LD_1;
	PORTB |= ledBuffer[0];
	waitms(t_on);
	PORTB &= ~LED_MASK;
	PORTD &= ~LD_1;

	waitms(t_off);
	PORTD |= LD_2;
	PORTB |= ledBuffer[1];
	waitms(t_on);
	PORTB &= ~LED_MASK;
	PORTD &= ~LD_2;

	waitms(t_off);
	PORTD |= LD_3;
	PORTB |= ledBuffer[2];
	waitms(t_on);
	PORTB &= ~LED_MASK;
	PORTD &= ~LD_3;
}

/* Precalculate hardware setup for display */
void display(char data[]) {
	for (uint8_t i = 0; i<3; i++) {
		ledBuffer[i] = segment_map[21];
		if (data[i] == '0') ledBuffer[i] = segment_map[0];
		if (data[i] == '1') ledBuffer[i] = segment_map[1];
		if (data[i] == '2') ledBuffer[i] = segment_map[2];
		if (data[i] == '3') ledBuffer[i] = segment_map[3];
		if (data[i] == '4') ledBuffer[i] = segment_map[4];
		if (data[i] == '5') ledBuffer[i] = segment_map[5];
		if (data[i] == '6') ledBuffer[i] = segment_map[6];
		if (data[i] == '7') ledBuffer[i] = segment_map[7];
		if (data[i] == '8') ledBuffer[i] = segment_map[8];
		if (data[i] == '9') ledBuffer[i] = segment_map[9];

		if (data[i] == 'A') ledBuffer[i] = segment_map[10];
		if (data[i] == 'B') ledBuffer[i] = segment_map[11];
		if (data[i] == 'C') ledBuffer[i] = segment_map[12];
		if (data[i] == 'D') ledBuffer[i] = segment_map[13];
		if (data[i] == 'E') ledBuffer[i] = segment_map[14];
		if (data[i] == 'F') ledBuffer[i] = segment_map[15];

		if (data[i] == 'P') ledBuffer[i] = segment_map[16];
		if (data[i] == 'R') ledBuffer[i] = segment_map[17];
		if (data[i] == 'Y') ledBuffer[i] = segment_map[18];
		if (data[i] == 'N') ledBuffer[i] = segment_map[19];
		if (data[i] == ' ') ledBuffer[i] = segment_map[20];
		if (data[i] == '_') ledBuffer[i] = segment_map[21];
	}
}

void waitms(uint16_t ms) {
  uint16_t i;
  for (i=0;i<ms;i++) _delay_ms(1);
}
